import angr
import claripy

from .errors import RegNotFoundException, RopException


def gadget_to_asmstring(project, gadget):
    code = project.loader.memory.load(gadget.addr,gadget.block_length)
    md = project.arch.capstone
    return "; ".join(["%s %s" %(i.mnemonic, i.op_str) for i in md.disasm(code,gadget.addr)])


def get_ast_dependency(ast):
    """
    ast must be created from a symbolic state where registers values are named "sreg_REG-"
    looks for registers that if we make the register symbolic then the ast becomes symbolic
    :param ast: the ast of which we are trying to analyze dependencies
    :return: A set of register names which affect the ast
    """
    dependencies = set()

    for var in ast.variables:
        if var.startswith("sreg_"):
            dependencies.add(var[5:].split("-")[0])
        else:
            return set()
    return dependencies


def get_ast_controllers(test_state, ast, reg_deps):
    """
    looks for registers that we can make symbolic then the ast can be "anything"
    :param test_state: the input state
    :param ast: the ast of which we are trying to analyze controllers
    :param reg_deps: All registers which it depends on
    :return: A set of register names which can control the ast
    """

    test_val = 0x4141414141414141 % (2 << test_state.arch.bits)

    controllers = []
    if not ast.symbolic:
        return controllers

    # make sure it can't be symbolic if all the registers are constrained
    constrained_copy = test_state.copy()
    for reg in reg_deps:
        if not constrained_copy.registers.load(reg).symbolic:
            continue
        constrained_copy.add_constraints(constrained_copy.registers.load(reg) == test_val)
    if len(constrained_copy.solver.eval_upto(ast, 2)) > 1:
        return controllers

    for reg in reg_deps:
        constrained_copy = test_state.copy()
        for r in [a for a in reg_deps if a != reg]:
            # for bp and registers that might be set
            if not constrained_copy.registers.load(r).symbolic:
                continue
            constrained_copy.add_constraints(constrained_copy.registers.load(r) == test_val)

        if unconstrained_check(constrained_copy, ast):
            controllers.append(reg)

    return controllers


def unconstrained_check(state, ast):
    """
    Attempts to check if an ast is completely unconstrained
    :param state: the state to use
    :param ast: the ast to check
    :return: True if the ast is probably completely unconstrained
    """
    size = ast.size()
    test_val_0 = 0x0
    test_val_1 = (1 << size) - 1
    test_val_2 = int("1010"*16, 2) % (1 << size)
    test_val_3 = int("0101"*16, 2) % (1 << size)
    # chars need to be able to be different
    test_val_4 = int(("1001"*2 + "1010"*2 + "1011"*2 + "1100"*2 + "1101"*2 + "1110"*2 + "1110"*2 + "0001"*2), 2) \
        % (1 << size)
    if not state.solver.satisfiable(extra_constraints=(ast == test_val_0,)):
        return False
    if not state.solver.satisfiable(extra_constraints=(ast == test_val_1,)):
        return False
    if not state.solver.satisfiable(extra_constraints=(ast == test_val_2,)):
        return False
    if not state.solver.satisfiable(extra_constraints=(ast == test_val_3,)):
        return False
    if not state.solver.satisfiable(extra_constraints=(ast == test_val_4,)):
        return False
    return True


def fast_unconstrained_check(state, ast):
    """
    Attempts to check if an ast has any common unreversable operations mul, div
    :param state: the state to use
    :param ast: the ast to check
    :return: True if the ast is probably unconstrained
    """
    good_ops = {"Extract", "BVS", "__add__", "__sub__", "Reverse"}
    if len(ast.variables) != 1:
        return unconstrained_check(state, ast)

    passes_prefilter = True
    for a in ast.recursive_children_asts:
        if a.op not in good_ops:
            passes_prefilter = False
    if ast.op not in good_ops:
        passes_prefilter = False

    if passes_prefilter:
        return True

    return unconstrained_check(state, ast)


def get_reg_name(arch, reg_offset):
    """
    :param reg_offset: Tries to find the name of a register given the offset in the registers.
    :return: The register name
    """
    # todo does this make sense
    if reg_offset is None:
        raise RegNotFoundException("register offset is None")

    original_offset = reg_offset
    while reg_offset >= 0 and reg_offset >= original_offset - arch.bytes:
        if reg_offset in arch.register_names:
            return arch.register_names[reg_offset]
        else:
            reg_offset -= 1
    raise RegNotFoundException("register %s not found" % str(original_offset))


# todo this doesn't work if there is a timeout
def _asts_must_be_equal(state, ast1, ast2):
    """
    :param state: the state to use for solving
    :param ast1: first ast
    :param ast2: second ast
    :return: True if the ast's must be equal
    """
    if state.solver.satisfiable(extra_constraints=(ast1 != ast2,)):
        return False
    return True


def make_initial_state(project, stack_length):
    """
    :return: an initial state with a symbolic stack and good options for rop
    """
    initial_state = project.factory.blank_state(
        add_options={angr.options.AVOID_MULTIVALUED_READS, angr.options.AVOID_MULTIVALUED_WRITES,
                     angr.options.NO_SYMBOLIC_JUMP_RESOLUTION, angr.options.CGC_NO_SYMBOLIC_RECEIVE_LENGTH,
                     angr.options.NO_SYMBOLIC_SYSCALL_RESOLUTION, angr.options.TRACK_ACTION_HISTORY},
        remove_options=angr.options.resilience | angr.options.simplification)
    initial_state.options.discard(angr.options.CGC_ZERO_FILL_UNCONSTRAINED_MEMORY)
    initial_state.options.update({angr.options.TRACK_REGISTER_ACTIONS, angr.options.TRACK_MEMORY_ACTIONS,
                                  angr.options.TRACK_JMP_ACTIONS, angr.options.TRACK_CONSTRAINT_ACTIONS})
    symbolic_stack = claripy.Concat(*[
        initial_state.solver.BVS(f"symbolic_stack_{i}", project.arch.bits) for i in range(stack_length)
    ])
    initial_state.memory.store(initial_state.regs.sp, symbolic_stack)
    if initial_state.arch.bp_offset != initial_state.arch.sp_offset:
        initial_state.regs.bp = initial_state.regs.sp + 20*initial_state.arch.bytes
    initial_state.solver._solver.timeout = 500  # only solve for half a second at most
    return initial_state


def make_symbolic_state(project, reg_list, stack_length=80):
    """
    converts an input state into a state with symbolic registers
    :return: the symbolic state
    """
    input_state = make_initial_state(project, stack_length)
    symbolic_state = input_state.copy()
    # overwrite all registers
    for reg in reg_list:
        symbolic_state.registers.store(reg, symbolic_state.solver.BVS("sreg_" + reg + "-", project.arch.bits))
    # restore sp
    symbolic_state.regs.sp = input_state.regs.sp
    # restore bp
    symbolic_state.regs.bp = input_state.regs.bp
    return symbolic_state


def make_reg_symbolic(state, reg):
    state.registers.store(reg,
    state.solver.BVS("sreg_" + reg + "-", state.arch.bits))


def step_to_unconstrained_successor(project, state, max_steps=2, allow_simprocedures=False):
    """
    steps up to two times to try to find an unconstrained successor
    :param state: the input state
    :param max_steps: maximum number of additional steps to try to get to an unconstrained state
    :return: a path at the unconstrained successor
    """
    try:
        # might only want to enable this option for arches / oses which don't care about bad syscall
        # nums
        state.options.add(angr.options.BYPASS_UNSUPPORTED_SYSCALL)

        succ = project.factory.successors(state)
        if len(succ.flat_successors) + len(succ.unconstrained_successors) != 1:
            raise RopException("Does not get to a single successor")
        if len(succ.flat_successors) == 1 and max_steps > 0:
            if not allow_simprocedures and project.is_hooked(succ.flat_successors[0].addr):
                # it cannot be a syscall as now syscalls are not explicitly hooked
                raise RopException("Skipping simprocedure")
            return step_to_unconstrained_successor(project, succ.flat_successors[0],
                                                   max_steps-1, allow_simprocedures)
        if len(succ.flat_successors) == 1 and max_steps == 0:
            raise RopException("Does not get to an unconstrained successor")
        return succ.unconstrained_successors[0]

    except (angr.errors.AngrError, angr.errors.SimError):
        raise RopException("Does not get to a single unconstrained successor")



class RopMemAccess(object):
    """Holds information about memory accesses
    Attributes:
        addr_dependencies (set): All the registers that affect the memory address.
        data_dependencies (set): All the registers that affect the data written.
        addr_constant (int): If the address is a constant it is stored here.
        data_constant (int): If the data is constant it is stored here.
        addr_size (int): Number of bits used for the address.
        data_size (int): Number of bits used for data
    """
    def __init__(self):
        self.addr_dependencies = set()
        self.addr_controllers = set()
        self.data_dependencies = set()
        self.data_controllers = set()
        self.addr_constant = None
        self.data_constant = None
        self.addr_size = None
        self.data_size = None
        self.op = None

    def __hash__(self):
        to_hash = sorted(self.addr_dependencies) + sorted(self.data_dependencies) + [self.addr_constant] + \
            [self.data_constant] + [self.addr_size] + [self.data_size]
        return hash(tuple(to_hash))

    def __eq__(self, other):
        if type(other) != RopMemAccess:
            return False
        if self.addr_dependencies != other.addr_dependencies or self.data_dependencies != other.data_dependencies:
            return False
        if self.addr_controllers != other.addr_controllers or self.data_controllers != other.data_controllers:
            return False
        if self.addr_constant != other.addr_constant or self.data_constant != other.data_constant:
            return False
        if self.addr_size != other.addr_size or self.data_size != other.data_size:
            return False
        return True


class RopRegMove(object):
    """
    Holds information about Register moves
    Attributes:
        from_reg (string): register that started with the data
        to_reg (string): register that the data was moved to
        bits (int): number of bits that were moved
    """
    def __init__(self, from_reg, to_reg, bits):
        self.from_reg = from_reg
        self.to_reg = to_reg
        self.bits = bits

    def __hash__(self):
        return hash((self.from_reg, self.to_reg, self.bits))

    def __eq__(self, other):
        if type(other) != RopRegMove:
            return False
        return self.from_reg == other.from_reg and self.to_reg == other.to_reg and self.bits == other.bits


class RopGadget(object):
    def __init__(self, addr):
        self.addr = addr
        self.changed_regs = set()
        self.popped_regs = set()
        self.reg_dependencies = dict()  # like rax might depend on rbx, rcx
        self.reg_controllers = dict()  # like rax might be able to be controlled by rbx (for any value of rcx)
        self.stack_change = None
        self.mem_reads = []
        self.mem_writes = []
        self.mem_changes = []
        self.reg_moves = []
        self.bp_moves_to_sp = None
        self.block_length = None
        self.makes_syscall = False
        self.starts_with_syscall = False

    def __str__(self):
        s = "Gadget %#x\n" % self.addr
        if self.bp_moves_to_sp:
            s += "Stack change: bp + %#x\n" % self.stack_change
        else:
            s += "Stack change: %#x\n" % self.stack_change
        s += "Changed registers: " + str(self.changed_regs) + "\n"
        s += "Popped registers: " + str(self.popped_regs) + "\n"
        for move in self.reg_moves:
            s += "Register move: [%s to %s, %d bits]\n" % (move.from_reg, move.to_reg, move.bits)
        s += "Register dependencies:\n"
        for reg in self.reg_dependencies:
            controllers = self.reg_controllers.get(reg, list())
            dependencies = [x for x in self.reg_dependencies[reg] if x not in controllers]
            s += "    " + reg + ": [" + " ".join(controllers) + " (" + " ".join(dependencies) + ")]" + "\n"
        for mem_access in self.mem_changes:
            if mem_access.op == "__add__":
                s += "Memory add:\n"
            elif mem_access.op == "__sub__":
                s += "Memory subtract:\n"
            elif mem_access.op == "__or__":
                s += "Memory or:\n"
            elif mem_access.op == "__and__":
                s += "Memory and:\n"
            else:
                s += "Memory change:\n"
            if mem_access.addr_constant is None:
                s += "    " + "address (%d bits) depends on: " % mem_access.addr_size
                s += str(list(mem_access.addr_dependencies)) + "\n"
            else:
                s += "    " + "address (%d bits): %#x\n" % (mem_access.addr_size, mem_access.addr_constant)
            s += "    " + "data (%d bits) depends on: " % mem_access.data_size
            s += str(list(mem_access.data_dependencies)) + "\n"
        for mem_access in self.mem_writes:
            s += "Memory write:\n"
            if mem_access.addr_constant is None:
                s += "    " + "address (%d bits) depends on: " % mem_access.addr_size
                s += str(list(mem_access.addr_dependencies)) + "\n"
            else:
                s += "    " + "address (%d bits): %#x\n" % (mem_access.addr_size, mem_access.addr_constant)
            if mem_access.data_constant is None:
                s += "    " + "data (%d bits) depends on: " % mem_access.data_size
                s += str(list(mem_access.data_dependencies)) + "\n"
            else:
                s += "    " + "data (%d bits): %#x\n" % (mem_access.data_size, mem_access.data_constant)
        for mem_access in self.mem_reads:
            s += "Memory read:\n"
            if mem_access.addr_constant is None:
                s += "    " + "address (%d bits) depends on: " % mem_access.addr_size
                s += str(list(mem_access.addr_dependencies)) + "\n"
            else:
                s += "    " + "address (%d bits): %#x" % (mem_access.addr_size, mem_access.addr_constant)
            s += "    " + "data (%d bits) stored in regs:" % mem_access.data_size
            s += str(list(mem_access.data_dependencies)) + "\n"
        if self.makes_syscall:
            s += "Makes a syscall\n"
        return s

    def __repr__(self):
        return "<Gadget %#x>" % self.addr

    def copy(self):
        out = RopGadget(self.addr)
        out.addr = self.addr
        out.changed_regs = set(self.changed_regs)
        out.popped_regs = set(self.popped_regs)
        out.reg_dependencies = dict(self.reg_dependencies)
        out.reg_controllers = dict(self.reg_controllers)
        out.stack_change = self.stack_change
        out.mem_reads = list(self.mem_reads)
        out.mem_changes = list(self.mem_changes)
        out.mem_writes = list(self.mem_writes)
        out.reg_moves = list(self.reg_moves)
        out.bp_moves_to_sp = self.bp_moves_to_sp
        out.block_length = self.block_length
        out.makes_syscall = self.makes_syscall
        out.starts_with_syscall = self.starts_with_syscall
        return out


class StackPivot(object):
    def __init__(self, addr):
        self.addr = addr
        self.sp_from_reg = None
        self.sp_popped_offset = None

    def __str__(self):
        s = "Pivot %#x\n" % self.addr
        if self.sp_from_reg is not None:
            s += "sp from reg: %s\n" % self.sp_from_reg
        elif self.sp_popped_offset is not None:
            s += "sp popped at %#x\n" % self.sp_popped_offset
        return s

    def __repr__(self):
        return "<Pivot %#x>" % self.addr



import setuptools
from distutils.core import setup

setup(
    name='angrop',
    version='9.0.6136',
    description='The rop chain builder based off of angr',
    packages=['angrop'],
    install_requires=[
        'progressbar',
        'angr==9.0.6136',
    ],
)



from angr.errors import SimEngineError, SimMemoryError
from angr.analyses.bindiff import differing_constants
from angr.analyses.bindiff import UnmatchedStatementsException
from angr import Analysis, register_analysis

from . import chain_builder
from . import gadget_analyzer
from . import common

import pickle
import inspect
import logging
import progressbar

from .errors import RopException
from .rop_gadget import RopGadget, StackPivot

from multiprocessing import Pool

l = logging.getLogger('angrop.rop')


_global_gadget_analyzer = None


# global initializer for multiprocessing
def _set_global_gadget_analyzer(rop_gadget_analyzer):
    global _global_gadget_analyzer
    _global_gadget_analyzer = rop_gadget_analyzer


def run_worker(addr):
    return _global_gadget_analyzer.analyze_gadget(addr)


# todo what if we have mov eax, [rsp+0x20]; ret (cache would need to know where it is or at least a min/max)
# todo what if we have pop eax; mov ebx, eax; need to encode that we cannot set them to different values
class ROP(Analysis):
    """
    This class is a semantic aware rop gadget finder
    It is a work in progress, so don't be surprised if something doesn't quite work

    After calling find_gadgets(), find_gadgets_single_threaded() or load_gadgets(),
    self.gadgets, self.stack_pivots, and self._duplicates is populated.
    Additionally, all public methods from ChainBuilder are copied into ROP.
    """

    def __init__(self, only_check_near_rets=True, max_block_size=20, max_sym_mem_accesses=4, fast_mode=None):
        """
        Initializes the rop gadget finder
        :param only_check_near_rets: If true we skip blocks that are not near rets
        :param max_block_size: limits the size of blocks considered, longer blocks are less likely to be good rop
                               gadgets so we limit the size we consider
        :param fast_mode: if set to True sets options to run fast, if set to False sets options to find more gadgets
                          if set to None makes a decision based on the size of the binary
        :return:
        """

        # params
        self._max_block_size = max_block_size
        self._only_check_near_rets = only_check_near_rets
        self._max_sym_mem_accesses = max_sym_mem_accesses

        a = self.project.arch
        self._sp_reg = a.register_names[a.sp_offset]
        self._ip_reg = a.register_names[a.ip_offset]
        self._base_pointer = a.register_names[a.bp_offset]

        # get list of multipurpose registers
        self._reg_list = a.default_symbolic_registers
        # prune the register list of the instruction pointer and the stack pointer
        self._reg_list = [r for r in self._reg_list if r not in (self._sp_reg, self._ip_reg)]

        # get ret locations
        self._ret_locations = self._get_ret_locations()

        # list of RopGadget's
        self.gadgets = []
        self.stack_pivots = []
        self._duplicates = []

        # RopChain settings
        self.badbytes = []
        self.roparg_filler = None

        num_to_check = self._num_addresses_to_check()
        # fast mode
        if fast_mode is None:
            if num_to_check > 20000:
                fast_mode = True
                l.warning("Enabling fast mode for large binary")
            else:
                fast_mode = False
        self._fast_mode = fast_mode

        if self._fast_mode:
            self._max_block_size = 12
            self._max_sym_mem_accesses = 1
            # Recalculate num addresses to check based on fast_mode settings
            num_to_check = self._num_addresses_to_check()

        l.info("There are %d addresses within %d bytes of a ret",
               num_to_check, self._max_block_size)

        # gadget analyzer
        self._gadget_analyzer = gadget_analyzer.GadgetAnalyzer(self.project, self._reg_list, self._max_block_size,
                                                               self._fast_mode, self._max_sym_mem_accesses)
        # chain builder
        self._chain_builder = None

        # silence annoying loggers
        logging.getLogger('angr.engines.vex.ccall').setLevel(logging.CRITICAL)
        logging.getLogger('angr.engines.vex.expressions.ccall').setLevel(logging.CRITICAL)
        logging.getLogger('angr.engines.vex.irop').setLevel(logging.CRITICAL)
        logging.getLogger('angr.state_plugins.symbolic_memory').setLevel(logging.CRITICAL)
        logging.getLogger('pyvex.lifting.libvex').setLevel(logging.CRITICAL)
        logging.getLogger('angr.procedures.cgc.deallocate').setLevel(logging.CRITICAL)

    def find_gadgets(self, processes=4, show_progress=True):
        """
        Finds all the gadgets in the binary by calling analyze_gadget on every address near a ret.
        Saves gadgets in self.gadgets
        Saves stack pivots in self.stack_pivots
        :param processes: number of processes to use
        """
        self.gadgets = []

        pool = Pool(processes=processes, initializer=_set_global_gadget_analyzer, initargs=(self._gadget_analyzer,))

        it = pool.imap_unordered(run_worker, self._addresses_to_check_with_caching(show_progress), chunksize=5)
        for gadget in it:
            if gadget is not None:
                if isinstance(gadget, RopGadget):
                    self.gadgets.append(gadget)
                elif isinstance(gadget, StackPivot):
                    self.stack_pivots.append(gadget)

        pool.close()

        # fix up gadgets from cache
        for g in self.gadgets:
            if g.addr in self._cache:
                dups = {g.addr}
                for addr in self._cache[g.addr]:
                    dups.add(addr)
                    g_copy = g.copy()
                    g_copy.addr = addr
                    self.gadgets.append(g_copy)
                self._duplicates.append(dups)
        self.gadgets = sorted(self.gadgets, key=lambda x: x.addr)
        self._reload_chain_funcs()

    def find_gadgets_single_threaded(self, show_progress=True):
        """
        Finds all the gadgets in the binary by calling analyze_gadget on every address near a ret
        Saves gadgets in self.gadgets
        Saves stack pivots in self.stack_pivots
        """
        self.gadgets = []

        _set_global_gadget_analyzer(self._gadget_analyzer)
        for _, addr in enumerate(self._addresses_to_check_with_caching(show_progress)):
            gadget = _global_gadget_analyzer.analyze_gadget(addr)
            if gadget is not None:
                if isinstance(gadget, RopGadget):
                    self.gadgets.append(gadget)
                elif isinstance(gadget, StackPivot):
                    self.stack_pivots.append(gadget)

        # fix up gadgets from cache
        for g in self.gadgets:
            if g.addr in self._cache:
                dups = {g.addr}
                for addr in self._cache[g.addr]:
                    dups.add(addr)
                    g_copy = g.copy()
                    g_copy.addr = addr
                    self.gadgets.append(g_copy)
                self._duplicates.append(dups)
        self.gadgets = sorted(self.gadgets, key=lambda x: x.addr)
        self._reload_chain_funcs()

    def save_gadgets(self, path):
        """
        Saves gadgets in a file.
        :param path: A path for a file where the gadgets are stored
        """
        with open(path, "wb") as f:
            pickle.dump(self._get_cache_tuple(), f)

    def load_gadgets(self, path):
        """
        Loads gadgets from a file.
        :param path: A path for a file where the gadgets are loaded
        """
        cache_tuple = pickle.load(open(path, "rb"))
        self._load_cache_tuple(cache_tuple)

    def set_badbytes(self, badbytes):
        """
        Define badbytes which should not appear in the generated ropchain.
        :param badbytes: a list of 8 bit integers
        """
        if not isinstance(badbytes, list):
            print("Require a list, e.g: [0x00, 0x09]")
            return
        self.badbytes = badbytes
        if len(self.gadgets) > 0:
            self.chain_builder._set_badbytes(self.badbytes)

    def set_roparg_filler(self, roparg_filler):
        """
        Define rop gadget filler argument. These will be used if the rop chain needs to pop
        useless registers.
        If roparg_filler is None, symbolic values will be used and the concrete values will
        be whatever the constraint solver chooses (usually 0).
        :param roparg_filler: A integer which is used when popping useless register or None.
        """
        if not isinstance(roparg_filler, (int, type(None))):
            print("Require an integer, e.g: 0x41414141 or None")
            return

        self.roparg_filler = roparg_filler
        if len(self.gadgets) > 0:
            self.chain_builder._set_roparg_filler(self.roparg_filler)

    def get_badbytes(self):
        """
        Returns list of badbytes.
        :returns the list of badbytes
        """
        return self.badbytes

    def _get_cache_tuple(self):
        return self.gadgets, self.stack_pivots, self._duplicates

    def _load_cache_tuple(self, cache_tuple):
        self.gadgets, self.stack_pivots, self._duplicates = cache_tuple
        self._reload_chain_funcs()

    def _reload_chain_funcs(self):
        for f_name, f in inspect.getmembers(self.chain_builder, predicate=inspect.ismethod):
            if f_name.startswith("_"):
                continue
            setattr(self, f_name, f)

    @property
    def chain_builder(self):
        if self._chain_builder is not None:
            return self._chain_builder
        elif len(self.gadgets) > 0:
            self._chain_builder = chain_builder.ChainBuilder(self.project, self.gadgets, self._duplicates,
                                                             self._reg_list, self._base_pointer, self.badbytes,
                                                             self.roparg_filler)
            return self._chain_builder
        else:
            raise Exception("No gadgets available, call find_gadgets() or load_gadgets() if you haven't already.")

    def _block_has_ip_relative(self, addr, bl):
        """
        Checks if a block has any ip relative instructions
        """
        string = bl.bytes
        test_addr = 0x41414140 + addr % 0x10
        bl2 = self.project.factory.block(test_addr, byte_string=string)
        try:
            diff_constants = differing_constants(bl, bl2)
        except UnmatchedStatementsException:
            return True
        # check if it changes if we move it
        bl_end = addr + bl.size
        bl2_end = test_addr + bl2.size
        filtered_diffs = []
        for d in diff_constants:
            if d.value_a < addr or d.value_a >= bl_end or \
                    d.value_b < test_addr or d.value_b >= bl2_end:
                filtered_diffs.append(d)
        return len(filtered_diffs) > 0

    def _addresses_to_check_with_caching(self, show_progress=True):
        num_addrs = self._num_addresses_to_check()
        widgets = ['ROP: ', progressbar.Percentage(), ' ',
                   progressbar.Bar(marker=progressbar.RotatingMarker()),
                   ' ', progressbar.ETA(), ' ', progressbar.FileTransferSpeed()]
        progress = progressbar.ProgressBar(widgets=widgets, maxval=num_addrs)
        if show_progress:
            progress.start()
        self._cache = dict()
        seen = dict()
        for i, a in enumerate(self._addresses_to_check()):
            if show_progress:
                progress.update(i)
            try:
                bl = self.project.factory.block(a)
                if bl.size > self._max_block_size:
                    continue
                block_data = bl.bytes
            except (SimEngineError, SimMemoryError):
                continue
            if block_data in seen:
                self._cache[seen[block_data]].add(a)
                continue
            else:
                if self._is_jumpkind_valid(bl.vex.jumpkind) and \
                        len(bl.vex.constant_jump_targets) == 0 and \
                        not self._block_has_ip_relative(a, bl):
                    seen[block_data] = a
                    self._cache[a] = set()
                yield a
        if show_progress:
            progress.finish()

    def _addresses_to_check(self):
        """
        :return: all the addresses to check
        """
        if self._only_check_near_rets:
            # align block size
            alignment = self.project.arch.instruction_alignment
            block_size = (self._max_block_size & ((1 << self.project.arch.bits) - alignment)) + alignment
            slices = [(addr-block_size, addr) for addr in self._ret_locations]
            current_addr = 0
            for st, _ in slices:
                current_addr = max(current_addr, st)
                end_addr = st + block_size + alignment
                for i in range(current_addr, end_addr, alignment):
                    segment = self.project.loader.main_object.find_segment_containing(i)
                    if segment is not None and segment.is_executable:
                        yield i
                current_addr = max(current_addr, end_addr)
        else:
            for segment in self.project.loader.main_object.segments:
                if segment.is_executable:
                    l.debug("Analyzing segment with address range: 0x%x, 0x%x" % (segment.min_addr, segment.max_addr))
                    for addr in range(segment.min_addr, segment.max_addr):
                        yield addr

    def _num_addresses_to_check(self):
        if self._only_check_near_rets:
            # TODO: This could probably be optimized further by fewer segments checks (i.e. iterating for segments and
            #  adding ranges instead of incrementing, instead of calling _addressses_to_check) although this is still a
            # significant improvement.
            return sum(1 for _ in self._addresses_to_check())
        else:
            num = 0
            for segment in self.project.loader.main_object.segments:
                if segment.is_executable:
                    num += (segment.max_addr - segment.min_addr)
            return num
                        
    def _get_ret_locations(self):
        """
        :return: all the locations in the binary with a ret instruction
        """

        try:
            return self._get_ret_locations_by_string()
        except RopException:
            pass

        addrs = []
        seen = set()
        for segment in self.project.loader.main_object.segments:
            if segment.is_executable:
                num_bytes = segment.max_addr-segment.min_addr

                alignment = self.project.arch.instruction_alignment
                # hack for arm thumb
                if self.project.arch.linux_name == "aarch64" or self.project.arch.linux_name == "arm":
                    alignment = 1

                # iterate through the code looking for rets
                for addr in range(segment.min_addr, segment.min_addr + num_bytes, alignment):
                    # dont recheck addresses we've seen before
                    if addr in seen:
                        continue
                    try:
                        block = self.project.factory.block(addr)
                        # it it has a ret get the return address
                        if block.vex.jumpkind.startswith("Ijk_Ret"):
                            ret_addr = block.instruction_addrs[-1]
                            # hack for mips pipelining
                            if self.project.arch.linux_name.startswith("mips"):
                                ret_addr = block.instruction_addrs[-2]
                            if ret_addr not in seen:
                                addrs.append(ret_addr)
                        # save the addresses in the block
                        seen.update(block.instruction_addrs)
                    except (SimEngineError, SimMemoryError):
                        pass

        return sorted(addrs)

    def _get_ret_locations_by_string(self):
        """
        uses a string filter to find the return instructions
        :return: all the locations in the binary with a ret instruction
        """
        if self.project.arch.linux_name == "x86_64" or self.project.arch.linux_name == "i386":
            ret_instructions = {b"\xc2", b"\xc3", b"\xca", b"\xcb"}
        else:
            raise RopException("Only have ret strings for i386 and x86_64")

        addrs = []
        try:
            for segment in self.project.loader.main_object.segments:
                if segment.is_executable:
                    num_bytes = segment.max_addr-segment.min_addr
                    read_bytes = self.project.loader.memory.load(segment.min_addr, num_bytes)
                    for ret_instruction in ret_instructions:
                        for loc in common.str_find_all(read_bytes, ret_instruction):
                            addrs.append(loc + segment.min_addr)
        except KeyError:
            l.warning("Key error with segment analysis")
            # try reading from state
            state = self.project.factory.entry_state()
            for segment in self.project.loader.main_object.segments:
                if segment.is_executable:
                    num_bytes = segment.max_addr - segment.min_addr

                    read_bytes = state.solver.eval(state.memory.load(segment.min_addr, num_bytes), cast_to=bytes)
                    for ret_instruction in ret_instructions:
                        for loc in common.str_find_all(read_bytes, ret_instruction):
                            addrs.append(loc + segment.min_addr)

        return sorted(addrs)

    @staticmethod
    def _is_jumpkind_valid(jk):

        if jk in {'Ijk_Boring', 'Ijk_Call', 'Ijk_Ret'}:
            return True
        return False


register_analysis(ROP, 'ROP')



import logging

import angr
import pyvex
import claripy
from angr.errors import SimEngineError, SimMemoryError

from . import rop_utils
from .rop_gadget import RopGadget, RopMemAccess, RopRegMove, StackPivot
from .errors import RopException, RegNotFoundException

l = logging.getLogger("angrop.gadget_analyzer")


class GadgetAnalyzer(object):
    def __init__(self, project, reg_list, max_block_size, fast_mode, max_sym_mem_accesses):
        # params
        self.project = project
        self._reg_list = reg_list
        self._max_block_size = max_block_size
        self._fast_mode = fast_mode
        self._max_sym_mem_accesses = max_sym_mem_accesses

        # initial state that others are based off
        self._stack_length = 80
        self._stack_length_bytes = self._stack_length * self.project.arch.bytes
        self._test_symbolic_state = rop_utils.make_symbolic_state(self.project, reg_list)
        self._stack_pointer_value = self._test_symbolic_state.solver.eval(self._test_symbolic_state.regs.sp)

        # architecture stuff
        self._base_pointer = self.project.arch.register_names[self.project.arch.bp_offset]
        self._sp_reg = self.project.arch.register_names[self.project.arch.sp_offset]

        # solve cache
        self._solve_cache = dict()

    def analyze_gadget(self, addr):
        """
        :param addr: address to analyze for a gadget
        :return: a RopGadget instance
        """
        l.info("Analyzing 0x%x", addr)

        # first check if the block makes sense
        if not self._block_makes_sense(addr):
            return None

        try:
            # unconstrained check prefilter
            if self._does_not_get_to_unconstrained(addr):
                l.debug("... does not get to unconstrained successor")
                return None

            # create the symbolic state at the address
            symbolic_state = self._test_symbolic_state.copy()
            symbolic_state.ip = addr
            symbolic_p = rop_utils.step_to_unconstrained_successor(self.project, state=symbolic_state)

            l.debug("... analyzing rop potential of block")

            # filter out those that dont get to a controlled successor
            l.info("... check for controlled successor")
            if not self._check_for_controlled_successor(symbolic_p, symbolic_state):
                pivot = self._check_pivot(symbolic_p, symbolic_state, addr)
                return pivot

            # filter out if too many mem accesses
            if not self._satisfies_mem_access_limits(symbolic_p):
                l.debug("... too many symbolic memory accesses")
                return None

            # create the gadget
            this_gadget = RopGadget(addr=addr)
            # FIXME this doesnt handle multiple steps
            this_gadget.block_length = self.project.factory.block(addr).size

            # compute sp change
            l.debug("... computing sp change")
            self._compute_sp_change(symbolic_state, this_gadget)

            if this_gadget.stack_change % (self.project.arch.bytes) != 0:
                l.debug("... uneven sp change")
                return None

            if this_gadget.stack_change <= 0:
                l.debug("stack change isn't positive")
                return None

            # if the sp moves to the bp we have to handle it differently
            if not this_gadget.bp_moves_to_sp and self._base_pointer != self._sp_reg:
                rop_utils.make_reg_symbolic(symbolic_state, self._base_pointer)
                symbolic_p = rop_utils.step_to_unconstrained_successor(self.project, symbolic_state)

                if not self._satisfies_mem_access_limits(symbolic_p):
                    l.debug("... too many symbolic memory accesses")
                    return None

            l.info("... checking for syscall availability")
            this_gadget.makes_syscall = self._does_syscall(symbolic_p)
            this_gadget.starts_with_syscall = self._starts_with_syscall(addr)

            l.info("... checking for controlled regs")
            self._check_reg_changes(symbolic_p, symbolic_state, this_gadget)

            # check for reg moves
            # get reg reads
            reg_reads = self._get_reg_reads(symbolic_p)
            l.debug("... checking for reg moves")
            self._check_reg_change_dependencies(symbolic_state, symbolic_p, this_gadget)
            self._check_reg_movers(symbolic_state, symbolic_p, reg_reads, this_gadget)

            # check mem accesses
            l.debug("... analyzing mem accesses")
            self._analyze_mem_accesses(symbolic_p, symbolic_state, this_gadget)
            for m_access in this_gadget.mem_writes + this_gadget.mem_reads + this_gadget.mem_changes:
                if len(m_access.addr_dependencies) == 0 and m_access.addr_constant is None:
                    l.debug("... mem access with no addr dependencies")
                    return None

        except RopException as e:
            l.debug("... %s", e)
            return None
        except claripy.ClaripyFrontendError as e:
            l.warning("... claripy error: %s", e)
            return None

        l.debug("... Appending gadget!")
        return this_gadget

    def _block_makes_sense(self, addr):
        """
        Checks if a block at addr makes sense to analyze for rop gadgets
        :param addr: the address to check
        :return: True or False
        """
        try:
            l.debug("... checking if block makes sense")
            block = self.project.factory.block(addr)

            # handle stuff that is incorrectly modelled
            if self.project.arch.linux_name == "x86_64" or self.project.arch.linux_name == "i386":
                capstr = str(block.capstone).lower()
                if 'cli' in capstr or 'rex' in capstr or "fs:" in capstr or "gs:" in capstr:
                    return False
                if block.size < 1 or block.bytes[0] == 0x4f:
                    return False

            if block.vex.jumpkind == 'Ijk_NoDecode':
                l.debug("... not decodable")
                return False

            if self._fast_mode:
                if block.vex.jumpkind != "Ijk_Ret" and not block.vex.jumpkind.startswith("Ijk_Sys"):
                    return False

            if any(isinstance(s, pyvex.IRStmt.Dirty) for s in block.vex.statements):
                l.debug("... has dirties that we probably can't handle")
                return False

            for op in block.vex.operations:
                if op.startswith("Iop_Div"):
                    return False

            if block.size > self._max_block_size:
                l.debug("... too long")
                return False

            # we don't like floating point stuff
            if "Ity_F16" in block.vex.tyenv.types or "Ity_F32" in block.vex.tyenv.types \
                    or "Ity_F64" in block.vex.tyenv.types or "Ity_F128" in block.vex.tyenv.types:
                return False

        except pyvex.PyVEXError:
            l.debug("... some pyvex")
            return False
        except (angr.AngrError, angr.SimCCallError):
            l.debug("... some other angr error")
            return False
        except angr.SimMemoryLimitError:
            l.debug("... angr memory limit error")
            return False
        except angr.UnsupportedIROpError:
            l.debug("... angr unsupported op error")
            return False
        except angr.UnsupportedSyscallError:
            return False
        except AttributeError:
            return False

        return True

    # todo this one skips syscalls so it doesnt need to step as far?
    def _does_not_get_to_unconstrained(self, addr, max_steps=2):
        try:
            # might miss jumps where one side is never satisfiable
            bl = self.project.factory.block(addr)
            constant_jump_targets = list(bl.vex.constant_jump_targets)
            if len(constant_jump_targets) == 1 and max_steps == 0:
                return True
            elif len(constant_jump_targets) == 1:
                if not self._block_makes_sense(constant_jump_targets[0]):
                    return True
                return self._does_not_get_to_unconstrained(constant_jump_targets[0], max_steps-1)
            elif len(constant_jump_targets) > 1:
                return True
            # 0 constant jump targets is what we want to find
            return False
        except (SimEngineError, SimMemoryError):
            return True

    def _satisfies_mem_access_limits(self, symbolic_path):
        """
        :param symbolic_path: the successor symbolic path
        :return: True/False indicating whether or not to keep the gadget
        """
        # get all the memory accesses
        symbolic_mem_accesses = []
        for a in reversed(symbolic_path.history.actions):
            if a.type == 'mem' and a.addr.ast.symbolic:
                symbolic_mem_accesses.append(a)
        if len(symbolic_mem_accesses) <= self._max_sym_mem_accesses:
            return True

        # allow mem changes (only add/subtract) to count as a single access
        if len(symbolic_mem_accesses) == 2 and self._max_sym_mem_accesses == 1:
            if symbolic_mem_accesses[0].action == "read" and symbolic_mem_accesses[1].action == "write" and \
                    (symbolic_mem_accesses[1].data.ast.op == "__sub__" or
                        symbolic_mem_accesses[1].data.ast.op == "__add__") and \
                    symbolic_mem_accesses[1].data.ast.size() == self.project.arch.bits and \
                    symbolic_mem_accesses[0].addr.ast is symbolic_mem_accesses[1].addr.ast:
                return True
        return False

    def _check_reg_changes(self, symbolic_p, symbolic_state, gadget):
        """
        Checks which registers were changed and which ones were popped
        :param symbolic_p: the stepped path, symbolic_state is an ancestor of it.
        :param symbolic_state: the input state for testing
        :param gadget: the gadget to store register change information
        """
        exit_target = symbolic_p.history.actions[-1].target.ast

        succ_state = symbolic_p
        stack_change = gadget.stack_change if not gadget.bp_moves_to_sp else None

        for reg in self._get_reg_writes(symbolic_p):
            # we assume any register in reg_writes changed
            # verify the stack controls it
            # we need to make sure they arent equal to the exit target otherwise they arent controlled
            # TODO what to do about moves to bp
            if symbolic_p.registers.load(reg) is exit_target:
                gadget.changed_regs.add(reg)
            elif self._check_if_stack_controls_ast(succ_state.registers.load(reg), symbolic_state, stack_change):
                gadget.popped_regs.add(reg)
                gadget.changed_regs.add(reg)
            else:
                gadget.changed_regs.add(reg)

    def _check_reg_change_dependencies(self, symbolic_state, symbolic_p, gadget):
        """
        Checks which registers affect register changes
        :param symbolic_state: the input state for testing
        :param symbolic_p: the stepped path, symbolic_state is an ancestor of it.
        :param gadget: the gadget to store the reg change dependencies in
        """
        for reg in gadget.changed_regs:
            # skip popped regs
            if reg in gadget.popped_regs:
                continue
            # check its dependencies
            dependencies = self._get_reg_dependencies(symbolic_p, reg)
            if len(dependencies) != 0:
                gadget.reg_dependencies[reg] = set(dependencies)
            controllers = self._get_reg_controllers(symbolic_state, symbolic_p, reg, dependencies)
            if len(controllers) != 0:
                gadget.reg_controllers[reg] = set(controllers)

    def _check_reg_movers(self, symbolic_state, symbolic_p, reg_reads, gadget):
        """
        Checks if any data is directly copied from one register to another
        :param symbolic_state: the input state for testing
        :param symbolic_p: the stepped path, symbolic_state is an ancestor of it.
        :param reg_reads: all the registers which were read
        :param gadget: the gadget in which to store the reg movers
        :return:
        """
        for reg in gadget.changed_regs:
            regs_to_check = reg_reads
            # skip popped regs
            if reg in gadget.popped_regs:
                continue
            # skip regs that depend on more than 1 reg
            if reg in gadget.reg_dependencies:
                if len(gadget.reg_dependencies[reg]) != 1:
                    continue
                regs_to_check = gadget.reg_dependencies[reg]
            for from_reg in regs_to_check:
                ast_1 = symbolic_state.registers.load(from_reg)
                ast_2 = symbolic_p.registers.load(reg)
                if ast_1 is ast_2:
                    gadget.reg_moves.append(RopRegMove(from_reg, reg, self.project.arch.bits))
                # try lower 32 bits (this is intended for amd64)
                # todo do this for less bits too?
                else:
                    half_bits = self.project.arch.bits // 2
                    ast_1 = claripy.Extract(half_bits-1, 0, ast_1)
                    ast_2 = claripy.Extract(half_bits-1, 0, ast_2)
                    if ast_1 is ast_2:
                        gadget.reg_moves.append(RopRegMove(from_reg, reg, half_bits))

    # todo need to handle reg calls/jumps
    def _check_for_controlled_successor(self, symbolic_p, symbolic_s):
        """
        :param symbolic_p: the symbolic successor path of symbolic_s
        :param symbolic_s: the symbolic state
        :return: True if the address of symbolic_p is controlled by the stack
        """
        return self._check_if_stack_controls_ast(symbolic_p.ip, symbolic_s)

    def _check_if_stack_controls_ast(self, ast, initial_state, gadget_stack_change=None):
        if gadget_stack_change is not None and gadget_stack_change <= 0:
            return False

        # if we had the lemma cache this might be already there!
        test_val = 0x4242424242424242 % (1 << self.project.arch.bits)

        # TODO add test where we recognize a value past the end of the stack frame isn't controlled
        # this is an annoying problem but this code should handle it

        # solve cache is used if it's already known to not work or
        # if we are using the whole stack (gadget_stack_change is None)
        if hash(ast) in self._solve_cache and \
                (gadget_stack_change is None or not self._solve_cache[hash(ast)]):
            return self._solve_cache[hash(ast)]

        # prefilter
        if len(ast.variables) != 1 or not list(ast.variables)[0].startswith("symbolic_stack"):
            self._solve_cache[hash(ast)] = False
            return False

        stack_bytes_length = self._stack_length * self.project.arch.bytes
        if gadget_stack_change is not None:
            stack_bytes_length = min(max(gadget_stack_change, 0), stack_bytes_length)
        concrete_stack = initial_state.solver.BVV(b"B" * stack_bytes_length)
        concrete_stack_s = initial_state.copy()
        concrete_stack_s.add_constraints(
            initial_state.memory.load(initial_state.regs.sp, stack_bytes_length) == concrete_stack)
        test_constraint = (ast != test_val)
        # stack must have set the register and it must be able to set the register to all 1's or all 0's
        if not concrete_stack_s.solver.satisfiable(extra_constraints=(test_constraint,)) and \
                rop_utils.fast_unconstrained_check(initial_state, ast):
            ans = True
        else:
            ans = False

        # only store the result if we were using the whole stack
        if gadget_stack_change is not None:
            self._solve_cache[hash(ast)] = ans
        return ans

    def _compute_sp_change(self, symbolic_state, gadget):
        """
        Computes the change in the stack pointer for a gadget, including whether or not it moves to the base pointer
        :param symbolic_state: the input symbolic state
        :param gadget: the gadget in which to store the sp change
        """
        # store symbolic sp and bp and check for dependencies
        ss_copy = symbolic_state.copy()
        ss_copy.regs.bp = ss_copy.solver.BVS("sreg_" + self._base_pointer + "-", self.project.arch.bits)
        ss_copy.regs.sp = ss_copy.solver.BVS("sreg_" + self._sp_reg + "-", self.project.arch.bits)
        symbolic_p = rop_utils.step_to_unconstrained_successor(self.project, ss_copy)
        dependencies = self._get_reg_dependencies(symbolic_p, "sp")
        sp_change = symbolic_p.regs.sp - ss_copy.regs.sp

        # analyze the results
        gadget.bp_moves_to_sp = False
        if len(dependencies) > 1:
            raise RopException("SP has multiple dependencies")
        elif len(dependencies) == 0 and sp_change.symbolic:
            raise RopException("SP change is uncontrolled")
        elif len(dependencies) == 0 and not sp_change.symbolic:
            stack_changes = [ss_copy.solver.eval(sp_change)]
        elif list(dependencies)[0] == self._sp_reg:
            stack_changes = ss_copy.solver.eval_upto(sp_change, 2)
            gadget.stack_change = stack_changes[0]
        elif list(dependencies)[0] == self._base_pointer:
            sp_change = symbolic_p.regs.sp - ss_copy.regs.bp
            stack_changes = ss_copy.solver.eval_upto(sp_change, 2)
            gadget.bp_moves_to_sp = True
        else:
            raise RopException("SP does not depend on SP or BP")

        if len(stack_changes) != 1:
            raise RopException("SP change is symbolic")

        gadget.stack_change = stack_changes[0]

    def _analyze_mem_accesses(self, symbolic_p, symbolic_state, gadget):
        """
        analyzes memory accesses and stores their info in the gadget
        :param symbolic_p: the stepped path, symbolic_state is an ancestor of it.
        :param symbolic_state: the input state for testing
        :param gadget: the gadget to store mem acccess in
        """
        last_action = None
        for a in symbolic_p.history.actions.hardcopy:
            if a.type == 'mem':
                mem_access = RopMemAccess()
                if a.addr.ast.symbolic:
                    mem_access.addr_dependencies = rop_utils.get_ast_dependency(a.addr.ast)
                    mem_access.addr_controllers = rop_utils.get_ast_controllers(symbolic_state, a.addr.ast,
                                                                                mem_access.addr_dependencies)
                else:
                    mem_access.addr_constant = symbolic_state.solver.eval(a.addr.ast)

                # don't need to inform user of stack reads/writes
                stack_min_addr = self._stack_pointer_value - 0x20
                # TODO should this be changed, so that we can more easily understand writes outside the frame
                stack_max_addr = max(stack_min_addr + self._stack_length_bytes, stack_min_addr + gadget.stack_change)
                if mem_access.addr_constant is not None and \
                        stack_min_addr <= mem_access.addr_constant < stack_max_addr:
                    continue

                if a.action == "write":
                    # special case for read than write form the same addr
                    if last_action is not None and last_action.action == "read" and \
                            last_action.addr.ast is a.addr.ast and \
                            last_action.ins_addr == a.ins_addr:
                        mem_change = gadget.mem_reads[-1]
                        gadget.mem_reads = gadget.mem_reads[:-1]
                        # get the actual change in certain cases
                        self._get_mem_change_op_and_data(mem_change, a, symbolic_state)
                        gadget.mem_changes.append(mem_change)
                        last_action = None
                        continue

                    # for writes we want what the data depends on
                    test_data = symbolic_state.solver.eval_upto(a.data.ast, 2)
                    if len(test_data) > 1:
                        mem_access.data_dependencies = rop_utils.get_ast_dependency(a.data.ast)
                        mem_access.data_controllers = rop_utils.get_ast_controllers(symbolic_state, a.data.ast,
                                                                                    mem_access.data_dependencies)
                    elif len(test_data) == 1:
                        mem_access.data_constant = test_data[0]
                    else:
                        raise Exception("No data values, something went wrong")
                elif a.action == "read" and not isinstance(a.data.ast, claripy.fp.FPV) and \
                        not isinstance(a.data.ast, claripy.ast.FP):
                    # for reads we want to know if any register will have the data after
                    succ_state = symbolic_p
                    bits_to_extend = self.project.arch.bits - a.data.ast.size()
                    # if bits_to_extend is negative it breaks everything, and we probably dont care about it
                    if bits_to_extend >= 0:
                        for reg in gadget.changed_regs:
                            # skip popped regs
                            if reg in gadget.popped_regs:
                                continue
                            # skip registers which have known dependencies
                            if reg in gadget.reg_dependencies and len(gadget.reg_dependencies[reg]) > 0:
                                continue
                            test_constraint = claripy.And(
                                succ_state.registers.load(reg) != a.data.ast.zero_extend(bits_to_extend),
                                succ_state.registers.load(reg) != a.data.ast.sign_extend(bits_to_extend))

                            if not succ_state.solver.satisfiable(extra_constraints=(test_constraint,)):
                                mem_access.data_dependencies.add(reg)

                mem_access.data_size = a.data.ast.size()
                mem_access.addr_size = a.addr.ast.size()

                last_action = a
                if a.action == "read":
                    gadget.mem_reads.append(mem_access)
                if a.action == "write":
                    gadget.mem_writes.append(mem_access)

    @staticmethod
    def _get_mem_change_op_and_data(mem_change, write_action, symbolic_state):
            if not write_action.data.ast.symbolic:
                return
            if len(write_action.data.ast.args) != 2:
                return
            if not write_action.data.ast.args[0].symbolic:
                return

            vars0 = list(write_action.data.ast.args[0].variables)
            if not len(vars0) == 1 and vars0[0].startswith("symbolic_read_sreg_"):
                return

            data_dependencies = rop_utils.get_ast_dependency(write_action.data.ast.args[1])
            if len(data_dependencies) != 1:
                return
            data_controllers = rop_utils.get_ast_controllers(symbolic_state, write_action.data.ast.args[1], data_dependencies)
            if len(data_controllers) != 1:
                return

            mem_change.op = write_action.data.ast.op
            mem_change.data_dependencies = data_dependencies
            mem_change.data_controllers = data_controllers

    def _does_syscall(self, symbolic_p):
        """
        checks if the path does a system call at some point
        :param symbolic_p: input path of which to check history
        """
        for addr in symbolic_p.history.bbl_addrs:
            if self.project.simos.is_syscall_addr(addr):
                return True

        return False

    def _check_pivot(self, symbolic_p, symbolic_state, addr):
        """
        Super basic pivot analysis. Pivots are not really used by angrop right now
        :param symbolic_p: the stepped path, symbolic_state is an ancestor of it.
        :param symbolic_state: input state for testing
        :return: the pivot object
        """
        if symbolic_p.history.depth > 1:
            return None
        pivot = None
        reg_deps = rop_utils.get_ast_dependency(symbolic_p.regs.sp)
        if len(reg_deps) == 1:
            pivot = StackPivot(addr)
            pivot.sp_from_reg = list(reg_deps)[0]
        elif len(symbolic_p.regs.sp.variables) == 1 and \
                list(symbolic_p.regs.sp.variables)[0].startswith("symbolic_stack"):
            offset = None
            for a in symbolic_p.regs.sp.recursive_children_asts:
                if a.op == "Extract" and a.depth == 2:
                    offset = a.args[2].size() - 1 - a.args[0]
            if offset is None or offset % 8 != 0:
                return None
            offset_bytes = offset//8
            pivot = StackPivot(addr)
            pivot.sp_popped_offset = offset_bytes

        if pivot is not None:
            # verify no weird mem accesses
            test_p = self.project.factory.simulation_manager(symbolic_state.copy())
            # step until we find the pivot action
            for i in range(self.project.factory.block(symbolic_state.addr).instructions):
                test_p.step(num_inst=1)
                if len(test_p.active) != 1:
                    return None
                if test_p.one_active.regs.sp.symbolic:
                    # found the pivot action
                    break
            # now iterate through the remaining instructions with a clean state
            test_p.step(num_inst=1)
            if len(test_p.active) != 1:
                return None
            succ1 = test_p.active[0]
            ss = symbolic_state.copy()
            ss.regs.ip = succ1.addr
            succ = self.project.factory.successors(ss)
            if len(succ.flat_successors + succ.unconstrained_successors) == 0:
                return None
            succ2 = (succ.flat_successors + succ.unconstrained_successors)[0]

            all_actions = succ1.history.actions.hardcopy + succ2.history.actions.hardcopy
            for a in all_actions:
                if a.type == "mem" and a.addr.ast.symbolic:
                    return None
            return pivot

        return None

    def _starts_with_syscall(self, addr):
        """
        checks if the path starts with a system call
        :param addr: input path to check history of
        """

        return self.project.factory.block(addr, num_inst=1).vex.jumpkind.startswith("Ijk_Sys")

    def _windup_to_presyscall_state(self, symbolic_p, symbolic_state):
        """
        Retrieve the state of a gadget just before the syscall is made
        :param symbolic_p: the stepped path, symbolic_state is an ancestor of it.
        :param symbolic_state: input state for testing
        """

        if self._does_syscall(symbolic_p):
            # step up until the syscall and save the possible syscall numbers into the gadget
            prev = cur = symbolic_state
            while not self._does_syscall(cur):
                succ = self.project.factory.successors(cur)
                prev = cur
                cur = succ.flat_successors[0]
            return prev.copy()

        raise RopException("Gadget passed to _windup_to_presyscall_state does not make a syscall")

    # UTILITY FUNCTIONS

    @staticmethod
    def _get_reg_dependencies(symbolic_p, test_reg):
        """
        Gets all the registers which affect a test register
        :param symbolic_p: the stepped path, symbolic_state is an ancestor of it.
        :param test_reg: the register of which we are trying to analyze dependencies
        :return: A set of register names which affect the test_reg
        """
        dependencies = rop_utils.get_ast_dependency(symbolic_p.registers.load(test_reg))
        return dependencies

    @staticmethod
    def _get_reg_controllers(symbolic_state, symbolic_p, test_reg, reg_deps):
        """
        Gets all the registers which can completely control a test register
        :param symbolic_state: the input state for testing
        :param symbolic_p: the stepped path, symbolic_state is an ancestor of it.
        :param test_reg: the register which we are trying to analyze controllers
        :param reg_deps: All registers which it depends on
        :return: A set of register names which can control the test_reg
        """
        controllers = rop_utils.get_ast_controllers(symbolic_state, symbolic_p.registers.load(test_reg), reg_deps)
        return controllers

    def _get_reg_reads(self, path):
        """
        Finds all the registers read in a path
        :param path: The path to check
        :return: A set of register names which are read
        """
        all_reg_reads = set()
        for a in reversed(path.history.actions):
            if a.type == "reg" and a.action == "read":
                try:
                    reg_name = rop_utils.get_reg_name(self.project.arch, a.offset)
                    if reg_name in self._reg_list:
                        all_reg_reads.add(reg_name)
                    elif reg_name != self._sp_reg:
                        l.info("reg read from register not in reg_list: %s", reg_name)
                except RegNotFoundException as e:
                    l.debug(e)
        return all_reg_reads

    def _get_reg_writes(self, path):
        """
        Finds all the registers written in a path
        :param path: The path to check
        :return: A set of register names which are read
        """
        all_reg_writes = set()
        for a in reversed(path.history.actions):
            if a.type == "reg" and a.action == "write":
                try:
                    reg_name = rop_utils.get_reg_name(self.project.arch, a.offset)
                    if reg_name in self._reg_list:
                        all_reg_writes.add(reg_name)
                    elif reg_name != self._sp_reg:
                        l.info("reg read from register not in reg_list: %s", reg_name)
                except RegNotFoundException as e:
                    l.debug(e)
        return all_reg_writes

# TODO ip setters, ie call rax
